use std::io;
use std::net::SocketAddr;
use std::time::Duration;
use std::sync::Arc;

use domain::base::iana::{Opcode, Rcode};
use domain::base::{message_builder, name, wire};
use domain::base::{MessageBuilder, Name, Rtype};
use domain::net::client::{tsig, stream};
use domain::tsig::{Algorithm, Key, KeyName};
use domain::net::client::request::{self, RequestMessage, RequestMessageMulti, SendRequest, SendRequestMulti};
use tokio::net::TcpStream;

use crate::resources::dns::internal;
use crate::proto;
use crate::errors::Error;
use super::{RecordDriver, ZoneDriver, DnsDriverError};

use async_trait::async_trait;

pub struct DnsDriverConfig {
    pub address: SocketAddr,
    pub tsig: Option<TsigConfig>
}

pub struct TsigConfig {
    pub key_name: KeyName,
    pub secret: Vec<u8>,
    pub algorithm: Algorithm
}

#[derive(Clone)]
pub struct DnsDriver {
    pub addr: SocketAddr,
    pub tsig_key: Option<Arc<Key>>,
}

type TsigDnsClient = tsig::Connection<stream::Connection<
    tsig::RequestMessage<request::RequestMessage<Vec<u8>>, Arc<Key>>,
    tsig::RequestMessage<request::RequestMessageMulti<Vec<u8>>, Arc<Key>>
>, Arc<Key>>;

impl DnsDriver {
    pub fn from_config(config: DnsDriverConfig) -> Self {
        let key = config.tsig.map(|tsig_config| {
            Arc::new(
                Key::new(
                    tsig_config.algorithm,
                    &tsig_config.secret,
                    tsig_config.key_name,
                    None,
                    None
                ).expect("Failed to build key"),
            )
        });

        Self {
            addr: config.address,
            tsig_key: key,
        }
    }

    async fn client<Req, ReqMulti>(&self) -> Result<stream::Connection<Req, ReqMulti>, DnsDriverError>
    where
        Req: request::ComposeRequest + Send + Sync + 'static,
        ReqMulti: request::ComposeRequestMulti + Send + Sync + 'static,
    {
        let mut stream_config = stream::Config::default();
        stream_config.set_response_timeout(
            Duration::from_millis(100),
        );
        let tcp_connect = TcpStream::connect(self.addr).await?;

        let (tcp_conn, transport) = stream::Connection::with_config(
            tcp_connect, stream_config
        );

        tokio::spawn(transport.run());

        Ok(tcp_conn)
    }

    async fn tsig_client(&self) -> Result<Option<TsigDnsClient>, DnsDriverError>
    {
        if let Some(ref key) = self.tsig_key {
            let conn = self.client().await?;
            Ok(Some(tsig::Connection::new(key.clone(), conn)))
        } else {
            Ok(None)
        }
    }
}

#[async_trait]
impl ZoneDriver for DnsDriver {
    async fn zone_exists(&self, zone: &str) -> Result<(), DnsDriverError> {
        let client = self.client::<_, RequestMessageMulti<Vec<u8>>>().await?;

        let mut msg = MessageBuilder::new_vec().question();
        msg.push((
            Name::vec_from_str(zone)?,
            Rtype::SOA,
        ))?;

        let req = RequestMessage::new(msg)?;

        let res = SendRequest::send_request(&client, req)
            .get_response()
            .await?;

        let rcode = res.header().rcode();

        match rcode {
            Rcode::NOERROR => Ok(()),
            Rcode::NXDOMAIN | Rcode::REFUSED => Err(DnsDriverError::ZoneNotFound {
                name: zone.to_string(),
            }),
            rcode => Err(DnsDriverError::ServerError {
                rcode: rcode.to_string(),
                name: zone.to_string(),
                qtype:  Rtype::SOA.to_string()
            })
        }
    }
}

#[async_trait]
impl RecordDriver for DnsDriver {
    /// ------------- AXFR -------------

    async fn get_records(&self, zone: &str) -> Result<internal::RecordList, DnsDriverError> {
        let mut msg = MessageBuilder::new_vec();
        msg.header_mut().set_ad(true);

        let mut msg = msg.question();
        msg.push((
            Name::vec_from_str(zone)?,
            Rtype::AXFR,
        ))?;

        let req = RequestMessageMulti::new(msg)?;

        let tsig_client = self.tsig_client().await?;

        let mut request = if let Some(client) = tsig_client {
            SendRequestMulti::send_request(&client, req)
        } else {
            let client = self.client::<RequestMessage<Vec<u8>>,_>().await?;
            SendRequestMulti::send_request(&client, req)
        };

        let mut records = Vec::new();

        while let Some(reply) = request.get_response().await? {
            let rcode = reply.header().rcode();

            if rcode != Rcode::NOERROR {
                return Err(DnsDriverError::ServerError {
                    rcode: rcode.to_string(),
                    name: zone.to_string(),
                    qtype:  Rtype::AXFR.to_string()
                });
            }

            let answer = reply.answer()?;

            for record in answer.limit_to::<proto::dns::ParsedRData<_, _>>() {
                let record = record?;
                records.push(record.into())
            }
        }

        // AXFR response ends with SOA, we remove it so it is not doubled in the response.
        records.pop();

        Ok(internal::RecordList { records })
    }

    /// ------------- Dynamic Update - RFC 2136 -------------
    ///
    /// 2 - Update Message Format
    ///   +---------------------+
    ///   |        Header       |
    ///   +---------------------+
    ///   |         Zone        | specifies the zone to be updated (RFC1035 Question)
    ///   +---------------------+
    ///   |     Prerequisite    | RRs or RRsets which must (not) preexist (RFC1035 Answer)
    ///   +---------------------+
    ///   |        Update       | RRs or RRsets to be added or deleted (RFC1035 Authority)
    ///   +---------------------+
    ///   |   Additional Data   | additional data
    ///   +---------------------+
    /// 2.2 - Message Header
    ///
    /// OPCODE is set to UPDATE.
    /// UPDATE uses only one flag bit (QR).
    ///
    /// 2.3 - Zone Section
    ///
    /// The ZNAME is the zone name, the ZTYPE must be SOA, and the ZCLASS is
    /// the zone's class.
    ///
    /// 3.2.4 - Table Of Metavalues Used In Prerequisite Section
    ///
    /// TTL must be specified as zero (0) for all prerequisite
    ///
    ///   CLASS    TYPE     RDATA    Meaning
    ///   ------------------------------------------------------------
    ///   ANY      ANY      empty    Name is in use
    ///   ANY      rrset    empty    RRset exists (value independent)
    ///   NONE     ANY      empty    Name is not in use
    ///   NONE     rrset    empty    RRset does not exist
    ///   zone     rrset    rr       RRset exists (value dependent) - Match against ALL RR in a RRset!!
    ///
    /// 3.4.2.6 - Table Of Metavalues Used In Update Section
    ///
    ///   CLASS    TYPE     RDATA    Meaning
    ///   ---------------------------------------------------------
    ///   ANY      ANY      empty    Delete all RRsets from a name - TTL must be specified as zero (0)
    ///   ANY      rrset    empty    Delete an RRset - TTL must be specified as zero (0)
    ///   NONE     rrset    rr       Delete an RR from an RRset - TTL must be specified as zero (0)
    ///   zone     rrset    rr       Add to an RRset


    async fn add_records(&self, zone: &str, new_records: internal::RecordList) -> Result<(), DnsDriverError> {
        let mut msg = MessageBuilder::new_vec();
        msg.header_mut().set_opcode(Opcode::UPDATE);

        let mut msg = msg.question();
        msg.push((
            Name::vec_from_str(zone)?,
            Rtype::SOA,
        ))?;

        let mut msg = msg.authority();

        for record in new_records.records {
            let record = proto::dns::RecordImpl::try_from(record)?;
            msg.push(record)?;
        }

        let req = RequestMessage::new(msg)?;

        let tsig_client = self.tsig_client().await?;

        let mut request = if let Some(client) = tsig_client {
            SendRequest::send_request(&client, req)
        } else {
            let client = self.client::<_, RequestMessageMulti<Vec<u8>>>().await?;
            SendRequest::send_request(&client, req)
        };

        let reply = request.get_response().await?;
        let rcode = reply.header().rcode();

        if rcode != Rcode::NOERROR {
            Err(DnsDriverError::ServerError {
                rcode: rcode.to_string(),
                name: zone.to_string(),
                qtype:  "UPDATE".to_string(),
            })
        } else {
            Ok(())
        }

    }
}

impl From<io::Error> for DnsDriverError {
    fn from(value: io::Error) -> Self {
        DnsDriverError::ConnectionError { reason: Box::new(value) }
    }
}

impl From<request::Error> for DnsDriverError {
    fn from(value: request::Error) -> Self {
        DnsDriverError::ConnectionError { reason: Box::new(value) }
    }
}

impl From<message_builder::PushError> for DnsDriverError {
    fn from(value: message_builder::PushError) -> Self {
        DnsDriverError::OperationError { reason: Box::new(value) }
    }
}

impl From<name::FromStrError> for DnsDriverError {
    fn from(value: name::FromStrError) -> Self {
        DnsDriverError::OperationError { reason: Box::new(value) }
    }
}

impl From<wire::ParseError> for DnsDriverError {
    fn from(value: wire::ParseError) -> Self {
        DnsDriverError::OperationError { reason: Box::new(value) }
    }
}

impl From<Error> for DnsDriverError {
    fn from(value: Error) -> Self {
        DnsDriverError::OperationError { reason: Box::new(value) }
    }
}

/*
use trust_dns_proto::DnsHandle;
use trust_dns_client::client::ClientHandle;
use trust_dns_client::rr::{DNSClass, RecordType};
use trust_dns_client::op::{UpdateMessage, OpCode, MessageType, Message, Query, ResponseCode, Edns};
use trust_dns_client::error::ClientError;

use super::{Name, Record, RData};
use super::client::{ClientResponse, DnsClient};
use super::connector::{RecordConnector, ZoneConnector, ConnectorError, ConnectorResult};


const MAX_PAYLOAD_LEN: u16 = 1232;


#[derive(Debug)]
pub enum DnsConnectorError {
    ClientError(ClientError),
    ResponceNotOk {
        code: ResponseCode,
        zone: Name,
    },
}

pub struct DnsConnectorClient {
    client: DnsClient
}

impl DnsConnectorClient {
    pub fn new(client: DnsClient) -> Self {
        DnsConnectorClient {
            client
        }
    }
}

impl ConnectorError for DnsConnectorError {
    fn zone_name(&self) -> Option<Name> {
        if let DnsConnectorError::ResponceNotOk { code: _code, zone } = self {
            Some(zone.clone())
        } else {
            None
        }
    }

    fn is_proto_error(&self) -> bool {
        return matches!(self, DnsConnectorError::ClientError(_));
    }
}

impl std::fmt::Display for DnsConnectorError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            DnsConnectorError::ClientError(e) => {
                write!(f, "DNS client error: {}", e)
            },
            DnsConnectorError::ResponceNotOk { code, zone } => {
                write!(f, "Query for zone \"{}\" failed with code \"{}\"", zone, code)
            }
        }

    }
}

fn set_edns(message: &mut Message) {
    let edns = message.extensions_mut().get_or_insert_with(Edns::new);
    edns.set_max_payload(MAX_PAYLOAD_LEN);
    edns.set_version(0);
}

#[async_trait]
impl RecordConnector for DnsConnectorClient {
    //type Error = DnsConnectorError;

    async fn get_records(&mut self, zone: Name, class: DNSClass) -> ConnectorResult<Vec<Record>>
    {
        let response = {
            let query = self.client.query(zone.clone(), class, RecordType::AXFR);
            match query.await.map_err(|e| Box::new(DnsConnectorError::ClientError(e))) {
                Err(e) => return Err(e),
                Ok(v) => v,
            }
        };

        if response.response_code() != ResponseCode::NoError {
            return Err(Box::new(DnsConnectorError::ResponceNotOk {
                code: response.response_code(),
                zone: zone,
            }));
        }

        let answers = response.answers();
        let mut records: Vec<_> = answers.to_vec().into_iter()
            .filter(|record| record.data().is_some() && !matches!(record.data().unwrap(), RData::NULL { .. } | RData::DNSSEC(_)))
            .collect();

        // AXFR response ends with SOA, we remove it so it is not doubled in the response.
        records.pop();
        Ok(records)
    }

    async fn add_records(&mut self, zone: Name, class: DNSClass, new_records: Vec<Record>) -> ConnectorResult<()>
    {
        // Taken from trust_dns_client::op::update_message::append
        // The original function can not be used as is because it takes a RecordSet and not a Record list

        let mut zone_query = Query::new();
        zone_query.set_name(zone.clone())
            .set_query_class(class)
            .set_query_type(RecordType::SOA);

        let mut message = Message::new();

        // TODO: set random / time based id
        message
            .set_id(0)
            .set_message_type(MessageType::Query)
            .set_op_code(OpCode::Update)
            .set_recursion_desired(false);
        message.add_zone(zone_query);
        message.add_updates(new_records);

        set_edns(&mut message);

        let response = match ClientResponse(self.client.send(message)).await.map_err(|e| Box::new(DnsConnectorError::ClientError(e))) {
            Err(e) => return Err(e),
            Ok(v) => v,
        };

        if response.response_code() != ResponseCode::NoError {
            return Err(Box::new(DnsConnectorError::ResponceNotOk {
                code: response.response_code(),
                zone: zone,
            }));
        }

        Ok(())
    }

    async fn update_records(&mut self, zone: Name, class: DNSClass, old_records: Vec<Record>, new_records: Vec<Record>) -> ConnectorResult<()>
    {
        // Taken from trust_dns_client::op::update_message::compare_and_swap
        // The original function can not be used as is because it takes a RecordSet and not a Record list

        // for updates, the query section is used for the zone
        let mut zone_query = Query::new();
        zone_query.set_name(zone.clone())
            .set_query_class(class)
            .set_query_type(RecordType::SOA);

        let mut message: Message = Message::new();

        // build the message
        // TODO: set random / time based id
        message
            .set_id(0)
            .set_message_type(MessageType::Query)
            .set_op_code(OpCode::Update)
            .set_recursion_desired(false);
        message.add_zone(zone_query);

        // make sure the record is what is expected
        let mut prerequisite = old_records.clone();
        for record in prerequisite.iter_mut() {
            record.set_ttl(0);
        }
        message.add_pre_requisites(prerequisite);

        // add the delete for the old record
        let mut delete = old_records;
        for record in delete.iter_mut() {
            // the class must be none for delete
            record.set_dns_class(DNSClass::NONE);
            // the TTL should be 0
            record.set_ttl(0);
        }
        message.add_updates(delete);

        // insert the new record...
        message.add_updates(new_records);

        // Extended dns
        set_edns(&mut message);

        let response = match ClientResponse(self.client.send(message)).await.map_err(|e| Box::new(DnsConnectorError::ClientError(e))) {
            Err(e) => return Err(e),
            Ok(v) => v,
        };

        if response.response_code() != ResponseCode::NoError {
            return Err(Box::new(DnsConnectorError::ResponceNotOk {
                code: response.response_code(),
                zone: zone,
            }));
        }

        Ok(())
    }

    async fn delete_records(&mut self, zone: Name, class: DNSClass, records: Vec<Record>) -> ConnectorResult<()>
    {
        // for updates, the query section is used for the zone
        let mut zone_query = Query::new();
        zone_query.set_name(zone.clone())
            .set_query_class(class)
            .set_query_type(RecordType::SOA);

        let mut message: Message = Message::new();

        // build the message
        // TODO: set random / time based id
        message
            .set_id(0)
            .set_message_type(MessageType::Query)
            .set_op_code(OpCode::Update)
            .set_recursion_desired(false);
        message.add_zone(zone_query);

        let mut delete = records;
        for record in delete.iter_mut() {
            // the class must be none for delete
            record.set_dns_class(DNSClass::NONE);
            // the TTL should be 0
            record.set_ttl(0);
        }
        message.add_updates(delete);

        // Extended dns
        set_edns(&mut message);

        let response = match ClientResponse(self.client.send(message)).await.map_err(|e| Box::new(DnsConnectorError::ClientError(e))) {
            Err(e) => return Err(e),
            Ok(v) => v,
        };

        if response.response_code() != ResponseCode::NoError {
            return Err(Box::new(DnsConnectorError::ResponceNotOk {
                code: response.response_code(),
                zone: zone,
            }));
        }

        Ok(())

    }
}


#[async_trait]
impl ZoneConnector for DnsConnectorClient {
    async fn zone_exists(&mut self, zone: Name, class: DNSClass) -> ConnectorResult<()>
    {
        let response = {
            info!("Querying SOA for name {}", zone);
            let query = self.client.query(zone.clone(), class, RecordType::SOA);
            match query.await.map_err(|e| Box::new(DnsConnectorError::ClientError(e))) {
                Err(e) => return Err(e),
                Ok(v) => v,
            }
        };

        if response.response_code() != ResponseCode::NoError {
            return Err(Box::new(DnsConnectorError::ResponceNotOk {
                code: response.response_code(),
                zone: zone,
            }));
        }

        Ok(())
    }

}
*/
