macro_rules! push_error {
    ($value:expr, $errors:expr) => {
        match $value {
            Err(error) => { $errors.push(error); None },
            Ok(value) => Some(value)
        }
    };
    ($value:expr, $errors:expr, $path:expr) => {
        match $value {
            Err(error) => { $errors.push(error.with_path($path)); None },
            Ok(value) => Some(value)
        }
    };
}

macro_rules! append_errors {
    ($value:expr, $errors:expr) => {
        match $value {
            Err(mut err) => { $errors.append(&mut err); None },
            Ok(value) => Some(value)
        }
    };
    ($value:expr, $errors:expr, $path:expr) => {
        match $value {
            Err(err) => { $errors.extend(err.into_iter().map(|e| {
                e.with_path($path)
            })); None },
            Ok(value) => Some(value)
        }
    };
}

macro_rules! check_type {
    ($value:expr, Number, $errors:expr) => {
        if let serde_json::Value::Number(value) = $value {
            Some(value)
        } else {

            let value = if let serde_json::Value::String(ref value) = $value {
                value.parse::<i64>().ok()
                    .map(serde_json::Number::from)
            } else {
                None
            };

            if value.is_some() {
                value
            } else {
                use crate::resources::dns::friendly::InputDataError;
                use crate::resources::dns::friendly::ValueType;
                use crate::errors::Error;

                $errors.push(Error::from(InputDataError::TypeError { expected: ValueType::Number, found: ValueType::from_value(&$value) }));
                None
            }
        }
    };
    ($value:expr, $type:ident, $errors:expr) => {
        if let serde_json::Value::$type(value) = $value {
            Some(value)
        } else {
            use crate::resources::dns::friendly::InputDataError;
            use crate::resources::dns::friendly::ValueType;
            use crate::errors::Error;

            $errors.push(Error::from(InputDataError::TypeError { expected: ValueType::$type, found: ValueType::from_value(&$value) }));
            None
        }
    };
}

macro_rules! get_object_value {
    ($value:expr, $key:expr, $errors:expr) => {
        if let Some(value) = $value.remove($key) {
            Some(value)
        } else {
            use crate::resources::dns::friendly::InputDataError;
            use crate::errors::Error;

            $errors.push(
                Error::from(InputDataError::MissingValue)
                    .with_path(concat!("/", $key))
            );
            None
        }
    };
}



pub(crate) use append_errors;
pub(crate) use push_error;
pub(crate) use check_type;
pub(crate) use get_object_value;
