use serde::{Deserialize, Serialize};

use crate::validation;
use crate::errors::Error;
use crate::macros::{append_errors, check_type, get_object_value, push_error};
use crate::resources::dns::internal::{self, Name, Record};

use super::{rdata, FriendlyRType};
use super::FromValue;


#[derive(Clone, Debug, Serialize, Deserialize)]
#[serde(rename_all = "lowercase")]
pub enum ConfigurationType {
    Mail,
    Web,
}

impl ConfigurationType {
    pub fn get_records(&self, input_data: serde_json::Value, default_ttl: u32, zone_name: Name) -> Result<Vec<Record>, Vec<Error>> {
        match self {
            ConfigurationType::Mail => {
                NewSectionMail::from_value(input_data.clone())
                    .map(|section| section.internal(default_ttl, zone_name))
            },
            ConfigurationType::Web => {
                NewSectionWeb::from_value(input_data.clone())
                    .map(|section| section.internal(default_ttl, zone_name))
            },
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
pub struct NewRecordQuery {
    pub name: Option<String>,
    pub config: Option<ConfigurationType>,
    pub rtype: Option<FriendlyRType>,
}

impl NewRecordQuery {
    pub fn validate(&mut self, zone_name: &str) -> Result<(), Vec<Error>> {
        let mut errors = Vec::new();
        if let Some(input_name) = &self.name {
            if input_name.is_empty() {
                self.name = Some(zone_name.to_string());
            } else {
                let new_name = format!("{input_name}.{zone_name}");
                let new_name = push_error!(
                    validation::normalize_domain(&new_name),
                    errors,
                    "/name"
                );
                // Keep the old value if the name is invalid
                if new_name.is_some() {
                    self.name = new_name;
                }
            }
        }

        if errors.is_empty() {
            Ok(())
        } else {
            Err(errors)
        }
    }
}

pub trait ToInternal {
    fn internal(self, ttl: u32, node_name: Name) -> Vec<internal::Record>;
}

#[derive(Debug)]
pub struct NewRecord<T> {
    ttl: Option<u32>,
    data: Option<T>,
}

impl<T: FromValue> FromValue for NewRecord<T> {
    fn from_value(value: serde_json::Value) -> Result<Self, Vec<crate::errors::Error>> {
        let mut errors = Vec::new();

        let object = check_type!(value, Object, errors);

        if !errors.is_empty() {
            return Err(errors);
        }

        let mut object = object.unwrap();

        let ttl = object.remove("ttl");
        let ttl = if let Some(ttl) = ttl {
            append_errors!(FromValue::from_value(ttl), errors, "/ttl")
        } else {
            None
        };

        let data = object.remove("data");
        let data = if let Some(data) = data {
            append_errors!(FromValue::from_value(data), errors, "/data")
        } else {
            None
        };

        if errors.is_empty() {
            Ok(NewRecord {
                ttl,
                data,
            })
        } else {
            Err(errors)
        }

    }
}

impl<T: ToInternal> ToInternal for NewRecord<T> {
    fn internal(self, ttl: u32, node_name: Name) -> Vec<internal::Record> {
        let ttl = self.ttl.unwrap_or(ttl);

        self.data
            .map(|data| data.internal(ttl, node_name))
            .unwrap_or_default()
    }
}

#[derive(Debug)]
pub struct NewRequiredRecord<T> {
    ttl: Option<u32>,
    data: T,
}

impl<T: FromValue> FromValue for NewRequiredRecord<T> {
    fn from_value(value: serde_json::Value) -> Result<Self, Vec<crate::errors::Error>> {
        let mut errors = Vec::new();

        let object = check_type!(value, Object, errors);

        if !errors.is_empty() {
            return Err(errors);
        }

        let mut object = object.unwrap();

        let ttl = object.remove("ttl");
        let ttl = if let Some(ttl) = ttl {
            append_errors!(FromValue::from_value(ttl), errors, "/ttl")
        } else {
            None
        };

        let data = get_object_value!(object, "data", errors);
        let data = if let Some(data) = data {
            append_errors!(FromValue::from_value(data), errors, "/data")
        } else {
            None
        };

        if errors.is_empty() {
            Ok(NewRequiredRecord {
                ttl,
                data: data.unwrap(),
            })
        } else {
            Err(errors)
        }

    }
}

impl<T: ToInternal> ToInternal for NewRequiredRecord<T> {
    fn internal(self, ttl: u32, node_name: Name) -> Vec<internal::Record> {
        let ttl = self.ttl.unwrap_or(ttl);

        self.data.internal(ttl, node_name)
    }
}

#[derive(Debug)]
pub struct NewSectionWeb {
    addresses: NewRequiredRecord<rdata::Addresses>,
}

impl FromValue for NewSectionWeb {
    fn from_value(value: serde_json::Value) -> Result<Self, Vec<Error>> {
        let mut errors = Vec::new();

        let object = check_type!(value, Object, errors);

        if !errors.is_empty() {
            return Err(errors);
        }

        let mut object = object.unwrap();

        let addresses = get_object_value!(object, "addresses", errors);
        let addresses = if let Some(addresses) = addresses {
            append_errors!(FromValue::from_value(addresses), errors, "/addresses")
        } else {
            None
        };

        if errors.is_empty() {
            Ok(NewSectionWeb {
                addresses: addresses.unwrap(),
            })
        } else {
            Err(errors)
        }

    }
}

impl ToInternal for NewSectionWeb {
    fn internal(self, ttl: u32, node_name: Name) -> Vec<internal::Record> {
        self.addresses.internal(ttl, node_name)
    }
}

#[derive(Debug)]
pub struct NewSectionMail {
    mailservers: NewRequiredRecord<rdata::MailServers>,
    spf: NewRecord<rdata::Spf>,
}

impl FromValue for NewSectionMail {
    fn from_value(value: serde_json::Value) -> Result<Self, Vec<Error>> {
        let mut errors = Vec::new();

        let object = check_type!(value, Object, errors);

        if !errors.is_empty() {
            return Err(errors);
        }

        let mut object = object.unwrap();

        let mailservers = get_object_value!(object, "mailservers", errors);
        let mailservers = if let Some(mailservers) = mailservers {
            append_errors!(FromValue::from_value(mailservers), errors, "/mailservers")
        } else {
            None
        };

        let spf = get_object_value!(object, "spf", errors);
        let spf = if let Some(spf) = spf {
            append_errors!(FromValue::from_value(spf), errors, "/spf")
        } else {
            None
        };


        if errors.is_empty() {
            Ok(NewSectionMail {
                mailservers: mailservers.unwrap(),
                spf: spf.unwrap()
            })
        } else {
            Err(errors)
        }

    }
}

impl ToInternal for NewSectionMail {
    fn internal(self, ttl: u32, node_name: Name) -> Vec<internal::Record> {
        let mut records =  self.mailservers.internal(ttl, node_name.clone());
        records.extend(self.spf.internal(ttl, node_name));
        records
    }
}
