use std::hash::Hash;
use std::net::IpAddr;

use serde::{Deserialize, Serialize};

use crate::errors::Error;
use crate::resources::dns::internal::{self, Name, Text};
use crate::macros::{append_errors, check_type, get_object_value};

use super::{FromValue, ToInternal};


#[derive(Debug, Deserialize, Serialize)]
pub enum FriendlyRData {
    Address(Address),
    Alias(Alias),
    MailServer(MailServer),
    NameServer(NameServer),
    Service(ServiceSingleTarget),
    Spf(Spf),
    TextData(TextData),
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(rename_all = "lowercase")]
pub enum FriendlyRDataAggregated {
    Addresses(Addresses),
    MailServers(MailServers),
    NameServers(NameServers),
    Service(Service),
    Spf(Spf),
    Texts(Texts),
}

/* --------- RDATA --------- */

/* --------- Address --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct Address {
    pub address: IpAddr,
}

impl FromValue for Address {
    fn from_value(value: serde_json::Value) -> Result<Self, Vec<Error>> {
        let mut errors = Vec::new();

        let object = check_type!(value, Object, errors);

        if !errors.is_empty() {
            return Err(errors);
        }

        let mut object = object.unwrap();

        let address = get_object_value!(object, "address", errors);
        let address = if let Some(address) = address {
            append_errors!(FromValue::from_value(address), errors, "/address")
        } else {
            None
        };

        if errors.is_empty() {
            Ok(Address {
                address: address.unwrap(),
            })
        } else {
            Err(errors)
        }
    }
}

#[derive(Debug, Deserialize, Serialize)]
pub struct Addresses {
    pub addresses: Vec<Address>,
}

impl FromValue for Addresses {
    fn from_value(value: serde_json::Value) -> Result<Self, Vec<Error>> {
        let mut errors = Vec::new();

        let object = check_type!(value, Object, errors);

        if !errors.is_empty() {
            return Err(errors);
        }

        let mut object = object.unwrap();

        let addresses = get_object_value!(object, "addresses", errors);
        let addresses = if let Some(addresses) = addresses {
            append_errors!(FromValue::from_value(addresses), errors, "/addresses")
        } else {
            None
        };

        if errors.is_empty() {
            Ok(Addresses {
                addresses: addresses.unwrap(),
            })
        } else {
            Err(errors)
        }

    }
}

impl ToInternal for Addresses {
    fn internal(self, ttl: u32, node_name: Name) -> Vec<internal::Record> {
        self.addresses.into_iter().map(|addr| {
            internal::Record {
                ttl,
                name: node_name.clone(),
                rdata: match addr.address {
                    IpAddr::V4(ip4) => internal::RData::A(internal::A::new(ip4)),
                    IpAddr::V6(ip6) => internal::RData::Aaaa(internal::Aaaa::new(ip6)),
                }
            }
        }).collect()
    }
}

/* --------- Service --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct ServiceSingleTarget {
    pub service_type: ServiceType,
    pub service_target: ServiceTarget,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct Service {
    pub service_type: ServiceType,
    pub service_targets: Vec<ServiceTarget>,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct ServiceTarget {
    pub port: i64,
    pub weight: i64,
    pub priority: i64,
    pub server: String,
}

#[derive(Debug, Hash, Eq, PartialEq, Clone, Deserialize, Serialize)]
#[serde(rename_all = "lowercase", tag = "service_type")]
pub enum ServiceType {
    Other { protocol: String, name: String },
}

/* --------- MailServer --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct MailServer {
    pub preference: u16,
    pub mail_exchanger: Name,
}

impl FromValue for MailServer {
    fn from_value(value: serde_json::Value) -> Result<Self, Vec<Error>> {
        let mut errors = Vec::new();

        let object = check_type!(value, Object, errors);

        if !errors.is_empty() {
            return Err(errors);
        }

        let mut object = object.unwrap();

        let preference = get_object_value!(object, "preference", errors);
        let preference = if let Some(preference) = preference {
            append_errors!(FromValue::from_value(preference), errors, "/preference")
        } else {
            None
        };

        let mail_exchanger = get_object_value!(object, "mail_exchanger", errors);
        let mail_exchanger = if let Some(mail_exchanger) = mail_exchanger {
            append_errors!(FromValue::from_value(mail_exchanger), errors, "/mail_exchanger")
        } else {
            None
        };

        if errors.is_empty() {
            Ok(MailServer {
                preference: preference.unwrap(),
                mail_exchanger: mail_exchanger.unwrap(),
            })
        } else {
            Err(errors)
        }
    }
}

#[derive(Debug, Deserialize, Serialize)]
pub struct MailServers {
    pub mailservers: Vec<MailServer>
}

impl FromValue for MailServers {
    fn from_value(value: serde_json::Value) -> Result<Self, Vec<Error>> {
        let mut errors = Vec::new();

        let object = check_type!(value, Object, errors);

        if !errors.is_empty() {
            return Err(errors);
        }

        let mut object = object.unwrap();

        let mailservers = get_object_value!(object, "mailservers", errors);
        let mailservers = if let Some(mailservers) = mailservers {
            append_errors!(FromValue::from_value(mailservers), errors, "/mailservers")
        } else {
            None
        };

        if errors.is_empty() {
            Ok(MailServers {
                mailservers: mailservers.unwrap(),
            })
        } else {
            Err(errors)
        }

    }
}

impl ToInternal for MailServers {
    fn internal(self, ttl: u32, node_name: Name) -> Vec<internal::Record> {
        self.mailservers.into_iter().map(|mailserver| {
            internal::Record {
                ttl,
                name: node_name.clone(),
                rdata: internal::RData::Mx(internal::Mx::new(mailserver.mail_exchanger, mailserver.preference)),
            }
        }).collect()
    }
}


/* --------- NameServer --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct NameServer {
    pub target: String,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct NameServers {
    pub nameservers: Vec<NameServer>
}

/* --------- TextData --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct TextData {
    pub text: String,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct Texts {
    pub texts: Vec<TextData>,
}

/* --------- Spf --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct Spf {
    pub policy: Text
}

impl FromValue for Spf {
    fn from_value(value: serde_json::Value) -> Result<Self, Vec<Error>> {
        let mut errors = Vec::new();

        let object = check_type!(value, Object, errors);

        if !errors.is_empty() {
            return Err(errors);
        }

        let mut object = object.unwrap();

        let policy = get_object_value!(object, "policy", errors);
        let policy = if let Some(policy) = policy {
            append_errors!(FromValue::from_value(policy), errors, "/policy")
        } else {
            None
        };

        if errors.is_empty() {
            Ok(Spf {
                policy: policy.unwrap(),
            })
        } else {
            Err(errors)
        }
    }
}


impl ToInternal for Spf {
    fn internal(self, ttl: u32, node_name: Name) -> Vec<internal::Record> {
        vec![
            internal::Record {
                ttl,
                name: node_name,
                rdata: internal::RData::Txt(internal::Txt::new(self.policy))
            }
        ]
    }
}

/* --------- Alias --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct Alias {
    pub from: String,
    pub target: String,
    pub ttl: i64,
}
