use std::collections::HashMap;
use std::hash::Hash;

use serde::{Serialize, Deserialize, Serializer};

use crate::resources::dns::internal;

use super::rdata::{
    FriendlyRDataAggregated,
    FriendlyRData,
    ServiceType,
    self
};


#[derive(Debug, Serialize)]
#[serde(rename_all="lowercase")]
pub enum RecordSection {
    Mail,
    Web,
    Services,
    Miscellaneous,
}


#[derive(Debug, Deserialize, Serialize, Hash, Eq, PartialEq)]
#[serde(rename_all="lowercase")]
pub enum FriendlyRType {
    Addresses,
    Alias,
    MailServers,
    NameServers,
    Service,
    Spf,
    Texts,
}

#[derive(Debug)]
pub struct FriendlyRecord {
    ttl: u32,
    data: FriendlyRDataAggregated,
}

impl Serialize for FriendlyRecord {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        #[derive(Serialize)]
        struct ExtendedRecord<'a> {
            ttl: u32,
            record_type: FriendlyRType,
            record_section: RecordSection,
            #[serde(flatten)]
            data: &'a FriendlyRDataAggregated,
        }

        let extended_record = ExtendedRecord {
            ttl: self.ttl,
            data: &self.data,
            record_type: self.record_type(),
            record_section: self.record_section(),
        };

        extended_record.serialize(serializer)

    }
}

impl FriendlyRecord {
    pub fn new(ttl: u32, data: FriendlyRDataAggregated) -> Self {
        FriendlyRecord {
            ttl,
            data,
        }
    }

    pub fn record_type(&self) -> FriendlyRType {
        match self.data {
            FriendlyRDataAggregated::Addresses(_) => FriendlyRType::Addresses,
            FriendlyRDataAggregated::MailServers(_) => FriendlyRType::MailServers,
            FriendlyRDataAggregated::NameServers(_) => FriendlyRType::NameServers,
            FriendlyRDataAggregated::Service(_) => FriendlyRType::Service,
            FriendlyRDataAggregated::Spf(_) => FriendlyRType::Spf,
            FriendlyRDataAggregated::Texts(_) => FriendlyRType::Texts,
        }
    }

    pub fn record_section(&self) -> RecordSection {
        match self.data {
            FriendlyRDataAggregated::Addresses(_) => RecordSection::Web,
            FriendlyRDataAggregated::MailServers(_) => RecordSection::Mail,
            FriendlyRDataAggregated::NameServers(_) => RecordSection::Miscellaneous,
            FriendlyRDataAggregated::Service(_) => RecordSection::Services,
            FriendlyRDataAggregated::Spf(_) => RecordSection::Mail,
            FriendlyRDataAggregated::Texts(_) => RecordSection::Miscellaneous,
        }
    }
}

#[derive(Debug, Serialize)]
pub struct Node {
    pub name: String,
    pub records: Vec<FriendlyRecord>,
}

impl Node {
    fn new(name: String) -> Self {
        Node {
            name,
            records: Vec::new(),
        }
    }
}

#[derive(Debug, Serialize)]
pub struct FriendlyRecords {
    records: Vec<Node>,
    aliases: Vec<rdata::Alias>,
}

impl FriendlyRecords {
    pub fn new() -> Self {
        FriendlyRecords {
            records: Vec::new(),
            aliases: Vec::new(),
        }
    }
}

impl From<internal::RecordList> for FriendlyRecords {
    fn from(value: internal::RecordList) -> Self {

        let mut records = FriendlyRecords::new();
        let mut name_mapping: HashMap<String, HashMap<FriendlyRType, FriendlyRecord>> = HashMap::new();
        let mut service_mapping: HashMap<(String, ServiceType), FriendlyRecord> = HashMap::new();

        for record in value.records {
            let internal::Record { name, ttl, rdata } = record;
            let name = name.to_string();
            let rdata = rdata.friendly(&name, ttl);

            if rdata.is_none() {
                continue;
            }

            let (name, rdata) = rdata.unwrap();

            if let FriendlyRData::Alias(alias) = rdata {
                records.aliases.push(alias)
            } else {
                let node = name_mapping.entry(name.clone()).or_default();

                match rdata {
                    FriendlyRData::Address(address) => {
                        let addresses = node.entry(FriendlyRType::Addresses).or_insert_with(|| {
                            FriendlyRecord::new(ttl, FriendlyRDataAggregated::Addresses(rdata::Addresses {
                                addresses: Vec::new()
                            }))
                        });

                        match addresses.data {
                            FriendlyRDataAggregated::Addresses(ref mut addresses) => addresses.addresses.push(address),
                            _ => unreachable!(),
                        };
                    },
                    FriendlyRData::MailServer(mailserver) => {
                        let mailservers = node.entry(FriendlyRType::MailServers).or_insert_with(|| {
                            FriendlyRecord::new(ttl, FriendlyRDataAggregated::MailServers(rdata::MailServers {
                                mailservers: Vec::new()
                            }))
                        });

                        match mailservers.data {
                            FriendlyRDataAggregated::MailServers(ref mut mailservers) => mailservers.mailservers.push(mailserver),
                            _ => unreachable!(),
                        };
                    },
                    FriendlyRData::Spf(spf) => {
                        node.insert(FriendlyRType::Spf, FriendlyRecord::new(ttl, FriendlyRDataAggregated::Spf(spf)));
                    },
                    FriendlyRData::Service(service_single) => {
                        let service = service_mapping.entry((name.clone(), service_single.service_type.clone()))
                            .or_insert_with(|| {
                                FriendlyRecord::new(ttl, FriendlyRDataAggregated::Service(rdata::Service {
                                    service_type: service_single.service_type,
                                    service_targets: Vec::new(),

                                }))
                            });

                        match service.data {
                            FriendlyRDataAggregated::Service(ref mut service) => service.service_targets.push(service_single.service_target),
                            _ => unreachable!(),
                        };
                    },
                    FriendlyRData::NameServer(nameserver) => {
                        // TODO: NS -> Skip if NS for zone (authority), create Delegation section with glue + DS for others (how to check if record is glue?)
                        let nameservers = node.entry(FriendlyRType::NameServers).or_insert_with(|| {
                            FriendlyRecord::new(ttl, FriendlyRDataAggregated::NameServers(rdata::NameServers {
                                nameservers: Vec::new()
                            }))
                        });

                        match nameservers.data {
                            FriendlyRDataAggregated::NameServers(ref mut nameservers) => nameservers.nameservers.push(nameserver),
                            _ => unreachable!(),
                        };
                    },
                    FriendlyRData::TextData(text) => {
                        let texts = node.entry(FriendlyRType::Texts).or_insert_with(|| {
                            FriendlyRecord::new(ttl, FriendlyRDataAggregated::Texts(rdata::Texts {
                                texts: Vec::new()
                            }))
                        });

                        match texts.data {
                            FriendlyRDataAggregated::Texts(ref mut texts) => texts.texts.push(text),
                            _ => unreachable!(),
                        };
                    },
                    FriendlyRData::Alias(_) => {},
                }
            }
        }

        let mut nodes: HashMap<String, Node> = HashMap::new();

        for ((name, _), service) in service_mapping {
            let node = nodes.entry(name.clone())
                .or_insert_with(|| Node::new(name));
            node.records.push(service);
        }

        for (name, node_records) in name_mapping {
            let node = nodes.entry(name.clone())
                .or_insert_with(|| Node::new(name));
            for (_, record) in node_records {
                node.records.push(record);
            }
        }

        records.records = nodes.into_values().collect();

        records.records.sort_by_key(|node| node.name.clone());

        records
    }
}
