use std::net::{Ipv4Addr, Ipv6Addr};

use super::{Name, Rtype, Text};

use crate::resources::dns::friendly;

#[derive(Clone, Debug)]
pub enum RData {
    A(A),
    Aaaa(Aaaa),
    Cname(Cname),
    Mx(Mx),
    Ns(Ns),
    Ptr(Ptr),
    Soa(Soa),
    Srv(Srv),
    Txt(Txt),
}

impl RData {
    pub fn rtype(&self) -> Rtype {
        match self {
            RData::A(_) => Rtype::A,
            RData::Aaaa(_) => Rtype::Aaaa,
            RData::Cname(_) => Rtype::Cname,
            RData::Mx(_) => Rtype::Mx,
            RData::Ns(_) => Rtype::Ns,
            RData::Ptr(_) => Rtype::Ptr,
            RData::Soa(_) => Rtype::Soa,
            RData::Srv(_) => Rtype::Srv,
            RData::Txt(_) => Rtype::Txt,
        }
    }

    pub fn friendly(self, owner: &str, ttl: u32) -> Option<(String, friendly::FriendlyRData)> {
        match self {
            RData::A(data) => data.friendly(owner),
            RData::Aaaa(data) => data.friendly(owner),
            RData::Cname(data) => data.friendly(owner, ttl),
            RData::Mx(data) => data.friendly(owner),
            RData::Ns(data) => data.friendly(owner),
            RData::Srv(data) => data.friendly(owner),
            _ => None,
        }
    }
}

#[derive(Clone, Debug)]
pub struct A {
    pub address: Ipv4Addr,
}

impl A {
    pub fn new(address: Ipv4Addr) -> Self {
        A {
            address
        }
    }

    pub fn friendly(self, owner: &str) -> Option<(String, friendly::FriendlyRData)> {
        Some((owner.to_string(), friendly::FriendlyRData::Address(friendly::Address {
            address: self.address.into()
        })))
    }
}

#[derive(Clone, Debug)]
pub struct Aaaa {
    pub address: Ipv6Addr,
}

impl Aaaa {

    pub fn new(address: Ipv6Addr) -> Self {
        Aaaa {
            address
        }
    }
    pub fn friendly(self, owner: &str) -> Option<(String, friendly::FriendlyRData)> {
        Some((owner.to_string(), friendly::FriendlyRData::Address(friendly::Address {
            address: self.address.into()
        })))
    }
}

#[derive(Clone, Debug)]
pub struct Cname {
    pub target: Name,
}

impl Cname {
    pub fn friendly(self, owner: &str, ttl: u32) -> Option<(String, friendly::FriendlyRData)> {
        Some((owner.to_string(), friendly::FriendlyRData::Alias(friendly::Alias {
            from: owner.into(),
            target: self.target.to_string(),
            ttl: ttl.into(),
        })))
    }
}

#[derive(Clone, Debug)]
pub struct Mx {
    pub preference: u16,
    pub mail_exchanger: Name,
}

impl Mx {
    pub fn new(mail_exchanger: Name, preference: u16) -> Self {
        Mx {
            mail_exchanger,
            preference,
        }
    }

    pub fn friendly(self, owner: &str) -> Option<(String, friendly::FriendlyRData)> {
        Some((owner.to_string(), friendly::FriendlyRData::MailServer(friendly::MailServer {
            preference: self.preference,
            mail_exchanger: self.mail_exchanger,
        })))
    }
}

#[derive(Clone, Debug)]
pub struct Ns {
    pub target: Name,
}

impl Ns {
    pub fn friendly(self, owner: &str) -> Option<(String, friendly::FriendlyRData)> {
        Some((owner.to_string(), friendly::FriendlyRData::NameServer(friendly::NameServer {
            target: self.target.to_string(),
        })))
    }
}

#[derive(Clone, Debug)]
pub struct Ptr {
    pub target: Name,
}

#[derive(Clone, Debug)]
pub struct Soa {
    pub primary_server: Name,
    pub maintainer: Name,
    pub refresh: u32,
    pub retry: u32,
    pub expire: u32,
    pub minimum: u32,
    pub serial: u32,
}

#[derive(Clone, Debug)]
pub struct Srv {
    pub server: Name,
    pub port: u16,
    pub priority: u16,
    pub weight: u16,
}

impl Srv {
    pub fn friendly(self, owner: &str) -> Option<(String, friendly::FriendlyRData)> {
        let labels: Vec<_> = owner.splitn(3, '.').collect();
        if labels.len() != 3 {
            None
        } else {
            let service_name = labels[0]. strip_prefix('_');
            let protocol = labels[1]. strip_prefix('_');
            if let (Some(service_name), Some(protocol)) = (service_name, protocol) {
                Some((labels[2].to_string(), friendly::FriendlyRData::Service(friendly::ServiceSingleTarget {
                    service_type: friendly::ServiceType::Other {
                        name: service_name.into(),
                        protocol: protocol.into()
                    },
                    service_target: friendly::ServiceTarget {
                        port: self.port.into(),
                        weight: self.weight.into(),
                        priority: self.priority.into(),
                        server: self.server.to_string()
                    }
                })))
            } else {
                None
            }
        }

    }
}

#[derive(Clone, Debug)]
pub struct Txt {
    pub text: Vec<u8>,
}

impl Txt {
    pub fn new(text: Text) -> Self {
        Txt {
            text: text.bytes()
        }
    }
}
