use axum::extract::{Query, Path, State, OriginalUri};
use axum::http::StatusCode;
use axum::response::{IntoResponse, Response};
use axum::Extension;
use serde_json::{Value, json};
use unic_langid::LanguageIdentifier;

use crate::form::Node;
use crate::macros::append_errors;
use crate::AppState;
use crate::errors::{Error, error_map};
use crate::template::Template;
use crate::resources::dns::friendly::{self, NewRecordQuery};
use crate::resources::dns::internal;

pub async fn get_records_page(
    Path(zone_name): Path<String>,
    State(app): State<AppState>,
    OriginalUri(url): OriginalUri,
    Extension(lang): Extension<LanguageIdentifier>,
) -> Result<Template<'static, Value>, Error> {
    let zone = app.db.get_zone_by_name(&zone_name).await?;
    let records = zone.get_records(app.records).await?;
    let records = friendly::FriendlyRecords::from(records.clone());

    Ok(Template::new(
        "pages/records.html",
        app.template_engine,
        json!({
            "current_zone": zone.name,
            "records": records,
            "url": url.to_string(),
            "lang": lang.to_string(),
        })
    ))
}


pub async fn get_new_record_page(
    Path(zone_name): Path<String>,
    State(app): State<AppState>,
    Query(mut params): Query<NewRecordQuery>,
    OriginalUri(url): OriginalUri,
    Extension(lang): Extension<LanguageIdentifier>,
) -> Result<Template<'static, Value>, Error> {
    let zone = app.db.get_zone_by_name(&zone_name).await?;

    let mut errors = Vec::new();
    append_errors!(params.validate(&zone.name), errors);

    Ok(Template::new(
        "pages/new_record.html",
        app.template_engine,
        json!({
            "current_zone": zone.name,
            "new_record_name": params.name,
            "errors": error_map(errors),
            "config": params.config,
            "rtype": params.rtype,
            "url": url.to_string(),
            "lang": lang.to_string(),
        })
    ))
}

pub async fn post_new_record(
    Path(zone_name): Path<String>,
    State(app): State<AppState>,
    Query(mut params): Query<NewRecordQuery>,
    OriginalUri(url): OriginalUri,
    Extension(lang): Extension<LanguageIdentifier>,
    form: Node,
) -> Result<Response, Error> {
    let zone = app.db.get_zone_by_name(&zone_name).await?;
    let mut errors = Vec::new();
    append_errors!(params.validate(&zone.name), errors);

    if !errors.is_empty() || params.name.is_none() || !(params.config.is_none() ^ params.config.is_some()) {
        return Ok(StatusCode::NOT_FOUND.into_response());
    }

    let name = params.name.clone().map(internal::Name::new).unwrap();
    let input_data = form.into_json_value();

    let new_records = {
        let new_records = if let Some(config_type) = params.config.clone() {
            config_type.get_records(input_data.clone(), 3600, name)
        } else if let Some(_rtype) = params.rtype {
            unimplemented!()
        } else {
            unreachable!()
        };

        append_errors!(new_records, errors)
    };

    if let Some(new_records) = new_records {
        zone.add_records(
            app.records,
            internal::RecordList::new(&new_records)
        ).await?;

        Ok(().into_response())
    } else {
        Ok(Template::new(
            "pages/new_record.html",
            app.template_engine,
            json!({
                "current_zone": zone.name,
                "new_record_name": params.name,
                "input_data": input_data,
                "errors": error_map(errors),
                "config": params.config,
                "rtype": params.rtype,
                "url": url.to_string(),
                "lang": lang.to_string(),
            })
        ).into_response())
    }

}
